@@ -19,6 +19,7 @@ use MIME::Base64;
 use Module::Find qw[useall];
 use Carp;
 use List::Util qw[max min sum];
+use POSIX ();
 
 use overload
   '""'  => \&string,
@@ -54,7 +55,7 @@ around 'new' => sub {
     my $self = shift;
 
     my $obj  = $self->$orig( @_ );
-    my $name = uc( '' . $obj->name );
+    my $name = lc( '' . $obj->name ) ;
     $name = '$$$NONAME' unless $name;
 
     if ( not exists $object_cache{$name}{ $obj->address->ip } ) {
@@ -176,12 +177,12 @@ sub query {
         } ## end if ( $name =~ m/(\.|^)\Q$fname\E$/i)
     } ## end foreach my $fname ( sort keys...)
 
-    if ( not exists( $self->cache->data->{"\U$name"}{"\U$type"}{"\U$class"}{$dnssec}{$usevc}{$recurse}{$edns_size} ) ) {
-        $self->cache->data->{"\U$name"}{"\U$type"}{"\U$class"}{$dnssec}{$usevc}{$recurse}{$edns_size} =
+    if ( not exists( $self->cache->data->{"$name"}{"\U$type"}{"\U$class"}{$dnssec}{$usevc}{$recurse}{$edns_size} ) ) {
+        $self->cache->data->{"$name"}{"\U$type"}{"\U$class"}{$dnssec}{$usevc}{$recurse}{$edns_size} =
           $self->_query( $name, $type, $href );
     }
 
-    my $p = $self->cache->data->{"\U$name"}{"\U$type"}{"\U$class"}{$dnssec}{$usevc}{$recurse}{$edns_size};
+    my $p = $self->cache->data->{"$name"}{"\U$type"}{"\U$class"}{$dnssec}{$usevc}{$recurse}{$edns_size};
     Zonemaster->logger->add( CACHED_RETURN => { packet => ( $p ? $p->string : 'undef' ) } );
 
     return $p;
@@ -319,6 +320,7 @@ sub compare {
 sub save {
     my ( $class, $filename ) = @_;
 
+    my $old = POSIX::setlocale(POSIX::LC_ALL,'C');
     my $json = JSON::XS->new->allow_blessed->convert_blessed;
     open my $fh, '>', $filename or die "Cache save failed: $!";
     foreach my $name ( keys %object_cache ) {
@@ -331,6 +333,7 @@ sub save {
 
     Zonemaster->logger->add( SAVED_NS_CACHE => { file => $filename } );
 
+    POSIX::setlocale(POSIX::LC_ALL, $old);
     return;
 }
 
@@ -406,7 +409,9 @@ sub median_time {
 
     my @t = sort { $a <=> $b } @{ $self->times };
     my $c = scalar( @t );
-    if ( $c % 2 == 0 ) {
+    if ( $c == 0 ) {
+        return 0;
+    } elsif ( $c % 2 == 0 ) {
         return ( $t[ $c / 2 ] + $t[ ( $c / 2 ) - 1 ] ) / 2;
     }
     else {
