@@ -178,6 +178,7 @@ sub metadata {
         dnssec03 => [
             qw(
               NO_NSEC3PARAM
+              NO_DNSKEY
               MANY_ITERATIONS
               TOO_MANY_ITERATIONS
               ITERATIONS_OK
@@ -185,6 +186,7 @@ sub metadata {
         ],
         dnssec04 => [
             qw(
+              RRSIG_EXPIRATION
               RRSIG_EXPIRED
               REMAINING_SHORT
               REMAINING_LONG
@@ -200,6 +202,7 @@ sub metadata {
               ALGORITHM_PRIVATE
               ALGORITHM_OK
               ALGORITHM_UNKNOWN
+              KEY_DETAILS
               )
         ],
         dnssec06 => [
@@ -293,6 +296,8 @@ sub translation {
 "RRSIG with keytag {tag} and covering type(s) {types} has a duration of {duration} seconds, which is too long.",
         "DURATION_OK" =>
 "RRSIG with keytag {tag} and covering type(s) {types} has a duration of {duration} seconds, which is just fine.",
+        "RRSIG_EXPIRATION" =>
+          "RRSIG with keytag {tag} and covering type(s) {types} expires at : {date}.",
         "RRSIG_EXPIRED" =>
           "RRSIG with keytag {tag} and covering type(s) {types} has already expired (expiration is: {expiration}).",
         "REMAINING_SHORT" =>
@@ -305,6 +310,7 @@ sub translation {
         "HAS_NSEC3"               => "The zone has NSEC3 records.",
         "INVALID_NAME_RCODE" => "When asked for the name {name}, which must not exist, the response had RCODE {rcode}.",
         "ITERATIONS_OK"      => "The number of NSEC3 iterations is {count}, which is OK.",
+        "KEY_DETAILS"        => "Key with keytag {keytag} details : Size = {keysize}, Flags ({sep}, {rfc5011}).",
         "MANY_ITERATIONS"    => "The number of NSEC3 iterations is {count}, which is on the high side.",
         "NEITHER_DNSKEY_NOR_DS" => "There are neither DS nor DNSKEY records for the zone.",
         "NO_COMMON_KEYTAGS"     => "No DS record had a DNSKEY with a matching keytag.",
@@ -366,6 +372,7 @@ sub policy {
         "HAS_NSEC3"                    => "INFO",
         "INVALID_NAME_RCODE"           => "NOTICE",
         "ITERATIONS_OK"                => "DEBUG",
+        "KEY_DETAILS"                  => "DEBUG",
         "MANY_ITERATIONS"              => "NOTICE",
         "NEITHER_DNSKEY_NOR_DS"        => "DEBUG",
         "NO_COMMON_KEYTAGS"            => "ERROR",
@@ -386,6 +393,7 @@ sub policy {
         "NSEC_SIG_VERIFY_ERROR"        => "ERROR",
         "REMAINING_LONG"               => "WARNING",
         "REMAINING_SHORT"              => "WARNING",
+        "RRSIG_EXPIRATION"             => "INFO",
         "RRSIG_EXPIRED"                => "ERROR",
         "SOA_NOT_SIGNED"               => "ERROR",
         "SOA_SIGNATURE_NOT_OK"         => "ERROR",
@@ -548,12 +556,6 @@ sub dnssec03 {
     my @nsec3params;
     @nsec3params = $param_p->get_records( 'NSEC3PARAM', 'answer' ) if $param_p;
 
-    my $dk_p = $zone->query_one( $zone->name, 'DNSKEY', { dnssec => 1 } );
-
-    my @dnskey;
-    @dnskey = $dk_p->get_records( 'DNSKEY', 'answer' ) if $dk_p;
-    my $min_len = min map { $_->keysize } @dnskey;
-
     if ( @nsec3params == 0 ) {
         push @results,
           info(
@@ -563,6 +565,30 @@ sub dnssec03 {
           );
     }
     else {
+        my $dk_p = $zone->query_one( $zone->name, 'DNSKEY', { dnssec => 1 } );
+
+        my @dnskey;
+        @dnskey = $dk_p->get_records( 'DNSKEY', 'answer' ) if $dk_p;
+
+        my $min_len = 0;
+        if ( @dnskey ) {
+            $min_len = min map { $_->keysize } @dnskey;
+            # Do rounding as per RFC5155 section 10.3
+            if ($min_len > 2048) {
+                $min_len = 4096;
+            }
+            elsif ($min_len > 1024) {
+                $min_len = 2048;
+            }
+            else {
+                $min_len = 1024;
+            }
+        }
+        else {
+            push @results,
+              info( NO_DNSKEY => {} );
+        }
+
         foreach my $n3p ( @nsec3params ) {
             my $iter = $n3p->iterations;
             if ( $iter > 100 ) {
@@ -585,7 +611,8 @@ sub dnssec03 {
                       );
                 }
             } ## end if ( $iter > 100 )
-            else {
+            elsif ( $min_len > 0 ) 
+            {
                 push @results,
                   info(
                     ITERATIONS_OK => {
@@ -620,6 +647,14 @@ sub dnssec04 {
     foreach my $sig ( @key_sigs, @soa_sigs ) {
         my $duration  = $sig->expiration - $sig->inception;
         my $remaining = $sig->expiration - int( $key_p->timestamp );
+        push @results,
+          info(
+            RRSIG_EXPIRATION => {
+                date  => scalar( gmtime($sig->expiration) ),
+                tag   => $sig->keytag,
+                types => $sig->typecovered,
+            }
+          );
         if ( $remaining < 0 ) {    # already expired
             push @results,
               info(
@@ -736,6 +771,17 @@ sub dnssec05 {
                     description => $algo_properties{$algo}{description},
                 }
               );
+            if ( $key->flags & 256 ) { # This is a Key
+                push @results,
+                  info(
+                    KEY_DETAILS => {
+                        keytag  => $key->keytag,
+                        keysize => $key->keysize,
+                        sep     => $key->flags & 1 ? q{SEP bit set} : q{SEP bit *not* set},
+                        rfc5011 => $key->flags & 128 ? q{RFC 5011 revocation bit set} : q{RFC 5011 revocation bit *not* set},
+                    }
+                );
+            }
         }
         else {
             push @results,
@@ -1021,7 +1067,10 @@ sub dnssec10 {
                 my $ok = 0;
                 foreach my $sig ( @sigs ) {
                     my $msg = q{};
-                    if (
+                    if (@dnskeys == 0) {
+                        push @results, info( NSEC_SIG_VERIFY_ERROR => { error => 'DNSKEY missing', sig => $sig->keytag } );
+                    }
+                    elsif (
                         $sig->verify_time(
                             [ grep { name( $_->name ) eq name( $sig->name ) } @nsec ],
                             \@dnskeys, $test_p->timestamp, $msg
